/*
 * Derived from checkpassword 0.71, beta, 19960530, which was
 * written by D. J. Bernstein, djb@pobox.com.
 * Changes by Bert Gijsbers (bert@mc.bio.uva.nl) to support
 * the timestamp argument and APOP authentication.
 * All changes donated to D. J. Bernstein.
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
extern int errno;
#include <pwd.h>
extern char *crypt();

/* configuration file */
#include "conf-apop.h"

/* prototypes from apopauth.c */
extern int apop_auth(/*char *loginname, char *passphrase, char *timestamp*/);
extern int has_apop(/*char *loginname*/);

static char up[512];
static int uplen;

void main(argc,argv)
int argc;
char **argv;
{
 char *login;
 char *password;
 char *timestamp;
 struct passwd *pw;
 char *encrypted;
 int r;
 int i;
 int apop_result;

 if (!argv[1]) _exit(2);

 uplen = 0;
 for (;;)
  {
   do
     r = read(3,up + uplen,sizeof(up) - uplen);
   while ((r == -1) && (errno == EINTR));
   if (r == -1) _exit(111);
   if (r == 0) break;
   uplen += r;
   if (uplen >= sizeof(up)) _exit(1);
  }

 close(3);

 i = 0;
 login = up + i;
 while (up[i++]) if (i == uplen) _exit(2);
 password = up + i;
 if (i == uplen) _exit(2);
 while (up[i++]) if (i == uplen) _exit(2);
 timestamp = 0;
 if (i < uplen)
  {
   timestamp = up + i;
   while (up[i++]) if (i == uplen) _exit(2);
  }

 if (login[0] == '/') _exit(1);

 if (!timestamp || !*timestamp)
  {
   /* RFC 1725 says that we must not allow oldstyle POP authentication
    * for a mailbox there is an APOP passphrase for it.
    */
   if (has_apop(login)) _exit(1);

   pw = getpwnam(login);
   /* XXX: unfortunately getpwnam() hides temporary errors */
   if (!pw) _exit(1);

   encrypted = crypt(password,pw->pw_passwd);

   for (i = 0;password[i];++i) password[i] = 0;

   if (!*pw->pw_passwd || strcmp(encrypted,pw->pw_passwd)) _exit(1);

   apop_result = 0;
  }
 else
  {
   apop_result = apop_auth(login, password, timestamp);

   for (i = 0;password[i];++i) password[i] = 0;

   if (apop_result <= 0 || apop_result > 2) _exit(1);
   if (apop_result == 2) pw = getpwnam(POPBOX_USERNAME);
   else pw = getpwnam(login);
   if (!pw) _exit(1);
  }

 if (setgid(pw->pw_gid) == -1) _exit(1);
 if (setuid(pw->pw_uid) == -1) _exit(1);
#ifdef CONF_MAILHOME
 if (chdir(CONF_MAILHOME) == -1) _exit(111);
 if (chdir(pw->pw_name) == -1) _exit(111);
#else
 if (chdir(pw->pw_dir) == -1) _exit(111);
#endif
 if (apop_result == 2)
  {
   for (i = 0; login[i]; i++)
    {
     if (login[i] == '.') login[i] = ':';
     if (login[i] == '/') _exit(1);
    }
   if (chdir(login) == -1) _exit(111);
  }

 strcpy(up,"USER=");
 strcat(up,pw->pw_name);
 if (putenv(up)) _exit(111);

 strcpy(up,"HOME=");
 strcat(up,pw->pw_dir);
 if (putenv(up)) _exit(111);

 strcpy(up,"SHELL=");
 strcat(up,pw->pw_shell);
 if (putenv(up)) _exit(111);

 execvp(argv[1],argv + 1);
 _exit(111);
}
